@@ -58,19 +58,24 @@ typedef enum {
     YELP_VIEW_STATE_ERROR
 } YelpViewState;
 
-GType              yelp_view_get_type          (void);
-GType              yelp_view_state_get_type    (void);
+GType              yelp_view_get_type             (void);
+GType              yelp_view_state_get_type       (void);
 
-GtkWidget *        yelp_view_new               (void);
-void               yelp_view_load              (YelpView      *view,
-                                                const gchar   *uri);
-void               yelp_view_load_uri          (YelpView      *view,
-                                                YelpUri       *uri);
-void               yelp_view_load_document     (YelpView      *view,
-                                                YelpUri       *uri,
-                                                YelpDocument  *document);
-YelpDocument *     yelp_view_get_document      (YelpView     *view);
-GtkActionGroup *   yelp_view_get_action_group  (YelpView     *view);
+GtkWidget *        yelp_view_new                  (void);
+void               yelp_view_load                 (YelpView        *view,
+                                                   const gchar     *uri);
+void               yelp_view_load_uri             (YelpView        *view,
+                                                   YelpUri         *uri);
+void               yelp_view_load_document        (YelpView        *view,
+                                                   YelpUri         *uri,
+                                                   YelpDocument    *document);
+YelpDocument *     yelp_view_get_document         (YelpView        *view);
+GtkActionGroup *   yelp_view_get_action_group     (YelpView        *view);
+
+void               yelp_view_add_link_action      (YelpView        *view,
+                                                   GtkAction       *action);
+YelpUri *          yelp_view_get_active_link_uri  (YelpView        *view);
+gchar *            yelp_view_get_active_link_text (YelpView        *view);
 
 G_END_DECLS
 
