@@ -54,6 +54,9 @@ static void          yelp_window_set_property     (GObject            *object,
 static void          window_construct             (YelpWindow         *window);
 static void          window_new                   (GtkAction          *action,
                                                    YelpWindow         *window);
+static gboolean      window_map_event             (YelpWindow         *window,
+                                                   GdkEvent           *event,
+                                                   gpointer            user_data);
 static gboolean      window_configure_event       (YelpWindow         *window,
                                                    GdkEventConfigure  *event,
                                                    gpointer            user_data);
@@ -70,6 +73,8 @@ static void          window_close                 (GtkAction          *action,
                                                    YelpWindow         *window);
 static void          window_add_bookmark          (GtkAction          *action,
                                                    YelpWindow         *window);
+static void          window_edit_bookmarks        (GtkAction          *action,
+                                                   YelpWindow         *window);
 static void          window_load_bookmark         (GtkAction          *action,
                                                    YelpWindow         *window);
 static void          window_find_in_page          (GtkAction          *action,
@@ -78,7 +83,13 @@ static void          window_start_search          (GtkAction          *action,
                                                    YelpWindow         *window);
 static void          window_open_location         (GtkAction          *action,
                                                    YelpWindow         *window);
-
+static void          window_read_later            (GtkAction          *action,
+                                                   YelpWindow         *window);
+static void          read_later_clicked           (GtkLinkButton      *button,
+                                                   YelpWindow         *window);
+static void          app_read_later_changed       (YelpApplication    *app,
+                                                   const gchar        *doc_uri,
+                                                   YelpWindow         *window);
 static void          app_bookmarks_changed        (YelpApplication    *app,
                                                    const gchar        *doc_uri,
                                                    YelpWindow         *window);
@@ -147,6 +158,24 @@ static gboolean      hidden_key_press             (GtkWidget          *widget,
                                                    GdkEventKey        *event,
                                                    YelpWindow         *window);
 
+static gboolean      bookmarks_closed             (GtkWindow          *bookmarks,
+                                                   GdkEvent           *event,
+                                                   YelpWindow         *window);
+static void          bookmarks_set_bookmarks      (YelpWindow         *window);
+static gboolean      bookmark_button_press        (GtkTreeView        *view,
+                                                   GdkEventButton     *event,
+                                                   YelpWindow         *window);
+static void          bookmark_opened              (GtkMenuItem        *item,
+                                                   YelpWindow         *window);
+static void          bookmark_activated           (GtkTreeView        *view,
+                                                   GtkTreePath        *path,
+                                                   GtkTreeViewColumn  *column,
+                                                   YelpWindow         *window);
+static gboolean      bookmark_key_release         (GtkTreeView        *view,
+                                                   GdkEventKey        *event,
+                                                   YelpWindow         *window);
+static void          bookmark_remove              (YelpWindow         *window);
+
 enum {
     PROP_0,
     PROP_APPLICATION
@@ -209,6 +238,7 @@ static const gchar *YELP_UI =
     "</menu>"
     "<menu action='BookmarksMenu'>"
     "<menuitem action='AddBookmark'/>"
+    "<menuitem action='EditBookmarks'/>"
     "<separator/>"
     "<placeholder name='Bookmarks'/>"
     "</menu>"
@@ -225,19 +255,27 @@ struct _YelpWindowPrivate {
     GtkActionGroup *action_group;
     YelpApplication *application;
 
+    GtkWidget    *bookmarks_editor;
+    /* no ref */
+    GtkWidget    *bookmarks_list;
+    GtkListStore *bookmarks_store;
+
     /* no refs on these, owned by containers */
     YelpView *view;
-    GtkWidget *vbox;
+    GtkWidget *vbox_view;
+    GtkWidget *vbox_full;
     GtkWidget *hbox;
     YelpLocationEntry *entry;
     GtkWidget *hidden_entry;
     GtkWidget *find_entry;
     GtkWidget *find_label;
+    GtkWidget *read_later_vbox;
 
     /* refs because we dynamically add & remove */
     GtkWidget *find_bar;
     GtkWidget *align_location;
     GtkWidget *align_hidden;
+    GtkWidget *read_later;
 
     gulong entry_location_selected;
 
@@ -257,6 +295,8 @@ struct _YelpWindowPrivate {
     gint find_cur_height;
     gint find_entry_height;
     gint find_bar_height;
+
+    gboolean configured;
 };
 
 static const GtkActionEntry entries[] = {
@@ -280,6 +320,11 @@ static const GtkActionEntry entries[] = {
       "<Control>D",
       NULL,
       G_CALLBACK (window_add_bookmark) },
+    { "EditBookmarks", NULL,
+      N_("_Edit Bookmarks"),
+      "<Control>B",
+      NULL,
+      G_CALLBACK (window_edit_bookmarks) },
     { "Find", NULL,
       N_("Find in Page..."),
       "<Control>F",
@@ -301,8 +346,11 @@ static void
 yelp_window_init (YelpWindow *window)
 {
     g_signal_connect (window, "configure-event", G_CALLBACK (window_configure_event), NULL);
+    g_signal_connect (window, "map-event", G_CALLBACK (window_map_event), NULL);
 }
 
+static void suppress_link_buttons (GtkLinkButton *button, const gchar *link, gpointer userdata) {}
+
 static void
 yelp_window_class_init (YelpWindowClass *klass)
 {
@@ -334,6 +382,13 @@ yelp_window_class_init (YelpWindowClass *klass)
                       G_TYPE_NONE, 0);
 
     g_type_class_add_private (klass, sizeof (YelpWindowPrivate));
+
+    /* We don't want GTK+'s default behavior for link buttons in the read
+     * later list. I don't see any other way around this, because GTK+'s
+     * handlers run before Yelp's. If we ever use link buttons elsewhere
+     * in Yelp, we need to do something in this callback.
+     */
+    gtk_link_button_set_uri_hook (suppress_link_buttons, NULL, NULL);
 }
 
 static void
@@ -381,6 +436,11 @@ yelp_window_dispose (GObject *object)
         priv->entry_color_animate = 0;
     }
 
+    if (priv->bookmarks_editor != NULL) {
+        gtk_widget_destroy (GTK_WIDGET (priv->bookmarks_editor));
+        priv->bookmarks_editor = NULL;
+    }
+
     G_OBJECT_CLASS (yelp_window_parent_class)->dispose (object);
 }
 
@@ -433,18 +493,27 @@ window_construct (YelpWindow *window)
     GtkWidget *scroll;
     GtkActionGroup *view_actions;
     GtkAction *action;
-    GtkWidget *button;
+    GtkWidget *vbox, *button;
     GtkTreeIter iter;
+    gchar *color, *text;
     YelpWindowPrivate *priv = GET_PRIV (window);
 
     gtk_window_set_icon_name (GTK_WINDOW (window), "help-browser");
 
     priv->view = (YelpView *) yelp_view_new ();
 
-    priv->vbox = gtk_vbox_new (FALSE, 0);
-    gtk_container_add (GTK_CONTAINER (window), priv->vbox);
+    action = gtk_action_new ("ReadLinkLater", "Read Link _Later", NULL, NULL);
+    g_signal_connect (action, "activate", G_CALLBACK (window_read_later), window);
+    yelp_view_add_link_action (priv->view, action);
+    g_signal_connect (priv->application, "read-later-changed", G_CALLBACK (app_read_later_changed), window);
 
-    priv->action_group = gtk_action_group_new ("WindowActions");
+    priv->vbox_full = gtk_vbox_new (FALSE, 3);
+    gtk_container_add (GTK_CONTAINER (window), priv->vbox_full);
+
+    priv->vbox_view = gtk_vbox_new (FALSE, 0);
+    gtk_box_pack_start (GTK_BOX (priv->vbox_full), priv->vbox_view, TRUE, TRUE, 0);
+
+    priv->action_group = gtk_action_group_new ("YelpWindowActions");
     gtk_action_group_set_translation_domain (priv->action_group, GETTEXT_PACKAGE);
     gtk_action_group_add_actions (priv->action_group,
 				  entries, G_N_ELEMENTS (entries),
@@ -464,7 +533,7 @@ window_construct (YelpWindow *window)
     gtk_window_add_accel_group (GTK_WINDOW (window),
                                 gtk_ui_manager_get_accel_group (priv->ui_manager));
     gtk_ui_manager_add_ui_from_string (priv->ui_manager, YELP_UI, -1, NULL);
-    gtk_box_pack_start (GTK_BOX (priv->vbox),
+    gtk_box_pack_start (GTK_BOX (priv->vbox_view),
                         gtk_ui_manager_get_widget (priv->ui_manager, "/ui/menubar"),
                         FALSE, FALSE, 0);
 
@@ -473,7 +542,7 @@ window_construct (YelpWindow *window)
 
     priv->hbox = gtk_hbox_new (FALSE, 0);
     g_object_set (priv->hbox, "border-width", 2, NULL);
-    gtk_box_pack_start (GTK_BOX (priv->vbox), priv->hbox, FALSE, FALSE, 0);
+    gtk_box_pack_start (GTK_BOX (priv->vbox_view), priv->hbox, FALSE, FALSE, 0);
 
     action = gtk_action_group_get_action (view_actions, "YelpViewGoBack");
     button = gtk_action_create_tool_item (action);
@@ -548,7 +617,7 @@ window_construct (YelpWindow *window)
                                     GTK_POLICY_AUTOMATIC);
     gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (scroll),
                                          GTK_SHADOW_IN);
-    gtk_box_pack_start (GTK_BOX (priv->vbox), scroll, TRUE, TRUE, 0);
+    gtk_box_pack_start (GTK_BOX (priv->vbox_view), scroll, TRUE, TRUE, 0);
 
     priv->find_bar = g_object_ref_sink (gtk_hbox_new (FALSE, 6));
     g_object_set (priv->find_bar, "border-width", 2, NULL);
@@ -570,6 +639,23 @@ window_construct (YelpWindow *window)
     g_object_set (priv->find_label, "xalign", 0.0, NULL);
     gtk_box_pack_start (GTK_BOX (priv->find_bar), priv->find_label, FALSE, FALSE, 0);
 
+    priv->read_later = g_object_ref_sink (gtk_info_bar_new ());
+    vbox = gtk_vbox_new (FALSE, 0);
+    color = yelp_settings_get_color (yelp_settings_get_default (),
+                                     YELP_SETTINGS_COLOR_TEXT_LIGHT);
+    text = g_markup_printf_escaped ("<span weight='bold' color='%s'>%s</span>",
+                                    color, _("Read Later"));
+    button = gtk_label_new (text);
+    g_object_set (button, "use-markup", TRUE, "xalign", 0.0, NULL);
+    g_free (color);
+    g_free (text);
+    gtk_box_pack_start (GTK_BOX (vbox), button, FALSE, FALSE, 0);
+    gtk_box_pack_start (GTK_BOX (gtk_info_bar_get_content_area (GTK_INFO_BAR (priv->read_later))),
+                        vbox,
+                        FALSE, FALSE, 0);
+    priv->read_later_vbox = gtk_vbox_new (FALSE, 0);
+    gtk_box_pack_start (GTK_BOX (vbox), priv->read_later_vbox, FALSE, FALSE, 0);
+
     g_signal_connect (priv->view, "external-uri", G_CALLBACK (view_external_uri), window);
     g_signal_connect (priv->view, "new-view-requested", G_CALLBACK (view_new_window), window);
     g_signal_connect (priv->view, "loaded", G_CALLBACK (view_loaded), window);
@@ -671,14 +757,38 @@ window_drag_received (YelpWindow         *window,
     gtk_drag_finish (context, FALSE, FALSE, time);
 }
 
+static gboolean
+window_map_event (YelpWindow  *window,
+                  GdkEvent    *event,
+                  gpointer     user_data)
+{
+    YelpWindowPrivate *priv = GET_PRIV (window);
+    priv->configured = TRUE;
+    return FALSE;
+}
+
 static gboolean
 window_configure_event (YelpWindow         *window,
                         GdkEventConfigure  *event,
                         gpointer            user_data)
 {
     YelpWindowPrivate *priv = GET_PRIV (window);
-    priv->width = event->width;
-    priv->height = event->height;
+    gboolean skip = TRUE;
+    if (priv->width != event->width) {
+        skip = FALSE;
+        priv->width = event->width;
+    }
+    if (priv->height != event->height) {
+        skip = FALSE;
+        priv->height = event->height;
+    }
+    /* Skip the configure-event signals that GTK+ sends as it's mapping
+     * the window, and also skip if the event didn't change the size of
+     * the window (i.e. it was just a move).
+     */
+    if (!priv->configured || skip)
+        return FALSE;
+
     if (priv->resize_signal > 0)
         g_source_remove (priv->resize_signal);
     priv->resize_signal = g_timeout_add (200,
@@ -728,6 +838,70 @@ window_add_bookmark (GtkAction  *action,
     g_object_unref (uri);
 }
 
+static void
+window_edit_bookmarks (GtkAction  *action,
+                       YelpWindow *window)
+{
+    YelpWindowPrivate *priv = GET_PRIV (window);
+    GtkWidget *scroll;
+    gchar *title;
+
+    if (priv->bookmarks_editor != NULL) {
+        gtk_window_present_with_time (GTK_WINDOW (priv->bookmarks_editor),
+                                      gtk_get_current_event_time ());
+        return;
+    }
+
+    priv->bookmarks_editor = gtk_window_new (GTK_WINDOW_TOPLEVEL);
+    gtk_window_set_transient_for (GTK_WINDOW (priv->bookmarks_editor),
+                                  GTK_WINDOW (window));
+    /* %s will be replaced with the name of a document */
+    title = g_strdup_printf (_("Bookmarks for %s"),
+                             gtk_window_get_title (GTK_WINDOW (window)));
+    gtk_window_set_title (GTK_WINDOW (priv->bookmarks_editor), title);
+    g_free (title);
+    gtk_container_set_border_width (GTK_CONTAINER (priv->bookmarks_editor), 6);
+    gtk_window_set_icon_name (GTK_WINDOW (priv->bookmarks_editor), "bookmark");
+    gtk_window_set_default_size (GTK_WINDOW (priv->bookmarks_editor), 300, 300);
+    g_signal_connect (priv->bookmarks_editor, "delete-event",
+                      G_CALLBACK (bookmarks_closed), window);
+
+    scroll = gtk_scrolled_window_new (NULL, NULL);
+    gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scroll),
+                                    GTK_POLICY_AUTOMATIC,
+                                    GTK_POLICY_AUTOMATIC);
+    gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (scroll),
+                                         GTK_SHADOW_IN);
+    gtk_container_add (GTK_CONTAINER (priv->bookmarks_editor), scroll);
+
+    priv->bookmarks_store = gtk_list_store_new (3, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING);
+    priv->bookmarks_list = gtk_tree_view_new_with_model (GTK_TREE_MODEL (priv->bookmarks_store));
+    gtk_container_add (GTK_CONTAINER (scroll), priv->bookmarks_list);
+
+    g_signal_connect (priv->bookmarks_list, "button-press-event",
+                      G_CALLBACK (bookmark_button_press), window);
+    g_signal_connect (priv->bookmarks_list, "row-activated",
+                      G_CALLBACK (bookmark_activated), window);
+    g_signal_connect (priv->bookmarks_list, "key-release-event",
+                      G_CALLBACK (bookmark_key_release), window);
+
+    gtk_tree_view_set_headers_visible (GTK_TREE_VIEW (priv->bookmarks_list), FALSE);
+    GtkCellRenderer *cell = gtk_cell_renderer_pixbuf_new ();
+    g_object_set (cell, "ypad", 2, "xpad", 2, NULL);
+    gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW (priv->bookmarks_list), 0, NULL,
+                                                 cell,
+                                                 "icon-name", 1,
+                                                 NULL);
+    gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW (priv->bookmarks_list), 1, NULL,
+                                                 gtk_cell_renderer_text_new (),
+                                                 "text", 2,
+                                                 NULL);
+
+    bookmarks_set_bookmarks (window);
+
+    gtk_widget_show_all (priv->bookmarks_editor);
+}
+
 static void
 window_load_bookmark (GtkAction  *action,
                       YelpWindow *window)
@@ -760,8 +934,11 @@ app_bookmarks_changed (YelpApplication *app,
     g_object_get (priv->view, "yelp-uri", &uri, NULL);
     this_doc_uri = yelp_uri_get_document_uri (uri);
 
-    if (g_str_equal (this_doc_uri, doc_uri))
+    if (g_str_equal (this_doc_uri, doc_uri)) {
         window_set_bookmarks (window, doc_uri);
+        if (priv->bookmarks_editor != NULL)
+            bookmarks_set_bookmarks (window);
+    }
 
     g_free (this_doc_uri);
     g_object_unref (uri);
@@ -812,6 +989,8 @@ window_set_bookmarks (YelpWindow  *window,
         entry->title = title;
         entries = g_slist_insert_sorted (entries, entry, (GCompareFunc) entry_compare);
     }
+    gtk_action_set_sensitive (gtk_action_group_get_action (priv->action_group, "EditBookmarks"),
+                              entries != NULL);
     for ( ; entries != NULL; entries = g_slist_delete_link (entries, entries)) {
         GSList *cur;
         GtkAction *bookmark;
@@ -819,7 +998,17 @@ window_set_bookmarks (YelpWindow  *window,
         gchar *action_id = g_strconcat ("LoadBookmark-", entry->page_id, NULL);
 
         bookmark = gtk_action_group_get_action (priv->bookmark_actions, action_id);
-        if (bookmark == NULL) {
+        if (bookmark) {
+            /* The action might have been set by a different document using
+             * the same page ID. We can just reuse the action, since it's
+             * just a page ID relative to the current URI, but we need to
+             * reset the title and icon.
+             */
+            g_object_set (bookmark,
+                          "label", entry->title,
+                          "icon-name", entry->icon,
+                          NULL);
+        } else {
             bookmark = gtk_action_new (action_id, entry->title, NULL, NULL);
             g_signal_connect (bookmark, "activate",
                               G_CALLBACK (window_load_bookmark), window);
@@ -1024,6 +1213,126 @@ window_open_location (GtkAction *action, YelpWindow *window)
     }
 }
 
+static void
+read_later_resolved (YelpUri    *uri,
+                     YelpWindow *window)
+{
+    gchar *fulluri;
+    const gchar *text = (const gchar *) g_object_get_data ((GObject *) uri, "link-text");
+    YelpWindowPrivate *priv = GET_PRIV (window);
+    YelpUri *base;
+    gchar *doc_uri;
+
+    g_object_get (priv->view, "yelp-uri", &base, NULL);
+    doc_uri = yelp_uri_get_document_uri (uri);
+    fulluri = yelp_uri_get_canonical_uri (uri);
+
+    yelp_application_add_read_later (priv->application, doc_uri, fulluri, text);
+
+    g_object_unref (base);
+    g_free (doc_uri);
+    g_free (fulluri);
+}
+
+static void
+window_read_later (GtkAction   *action,
+                   YelpWindow  *window)
+{
+    YelpWindowPrivate *priv = GET_PRIV (window);
+    YelpUri *uri;
+    gchar *text;
+
+    uri = yelp_view_get_active_link_uri (priv->view);
+    text = yelp_view_get_active_link_text (priv->view);
+
+    g_object_set_data_full ((GObject *) uri, "link-text", text, g_free);
+
+    if (!yelp_uri_is_resolved (uri)) {
+        g_signal_connect (uri, "resolved",
+                          G_CALLBACK (read_later_resolved),
+                          window);
+        yelp_uri_resolve (uri);
+    }
+    else {
+        read_later_resolved (uri, window);
+    }
+}
+
+static void
+read_later_clicked (GtkLinkButton  *button,
+                    YelpWindow     *window)
+{
+    YelpWindowPrivate *priv = GET_PRIV (window);
+    YelpUri *base;
+    gchar *doc_uri;
+    gchar *fulluri;
+
+    fulluri = g_strdup (gtk_link_button_get_uri (button));
+
+    g_object_get (priv->view, "yelp-uri", &base, NULL);
+    doc_uri = yelp_uri_get_document_uri (base);
+
+    yelp_application_remove_read_later (priv->application, doc_uri, fulluri);
+
+    g_object_unref (base);
+    g_free (doc_uri);
+
+    yelp_view_load (priv->view, fulluri);
+
+    g_free (fulluri);
+}
+
+static void
+app_read_later_changed (YelpApplication *app,
+                        const gchar     *doc_uri,
+                        YelpWindow      *window)
+{
+    GVariant *value;
+    GVariantIter *viter;
+    gchar *uri, *title; /* do not free */
+    GList *children;
+    gboolean has_children = FALSE;
+    YelpWindowPrivate *priv = GET_PRIV (window);
+
+    children = gtk_container_get_children (GTK_CONTAINER (priv->read_later_vbox));
+    while (children) {
+        gtk_container_remove (GTK_CONTAINER (priv->read_later_vbox),
+                              GTK_WIDGET (children->data));
+        children = g_list_delete_link (children, children);
+    }
+
+    value = yelp_application_get_read_later (priv->application, doc_uri);
+    g_variant_get (value, "a(ss)", &viter);
+    while (g_variant_iter_loop (viter, "(&s&s)", &uri, &title)) {
+        GtkWidget *align, *link;
+
+        align = gtk_alignment_new (0.0, 0.0, 0.0, 0.0);
+        g_object_set (align, "left-padding", 6, NULL);
+        gtk_box_pack_start (GTK_BOX (priv->read_later_vbox), align, FALSE, FALSE, 0);
+
+        link = gtk_link_button_new_with_label (uri, title);
+        g_object_set (link, "xalign", 0.0, NULL);
+        g_signal_connect (link, "clicked", G_CALLBACK (read_later_clicked), window);
+        gtk_container_add (GTK_CONTAINER (align), link);
+
+        gtk_widget_show_all (align);
+        has_children = TRUE;
+    }
+    g_variant_iter_free (viter);
+    g_variant_unref (value);
+
+    if (has_children) {
+        if (gtk_widget_get_parent (priv->read_later) == NULL) {
+            gtk_box_pack_end (GTK_BOX (priv->vbox_full), priv->read_later, FALSE, FALSE, 0);
+            gtk_widget_show_all (priv->read_later);
+        }
+    }
+    else {
+        if (gtk_widget_get_parent (priv->read_later) != NULL)
+            gtk_container_remove (GTK_CONTAINER (priv->vbox_full), priv->read_later);
+    }
+}
+
 static gboolean
 find_animate_open (YelpWindow *window) {
     YelpWindowPrivate *priv = GET_PRIV (window);
@@ -1057,7 +1366,7 @@ find_animate_close (YelpWindow *window) {
         priv->find_cur_height = 0;
         g_object_set (priv->find_bar, "height-request", -1, NULL);
         g_object_set (priv->find_entry, "height-request", -1, NULL);
-        gtk_container_remove (GTK_CONTAINER (priv->vbox), priv->find_bar); 
+        gtk_container_remove (GTK_CONTAINER (priv->vbox_view), priv->find_bar); 
         priv->find_animate = 0;
         return FALSE;
     }
@@ -1091,7 +1400,7 @@ window_find_in_page (GtkAction  *action,
 
     g_object_set (priv->find_entry, "width-request", 2 * priv->width / 3, NULL);
 
-    gtk_box_pack_end (GTK_BOX (priv->vbox), priv->find_bar, FALSE, FALSE, 0);
+    gtk_box_pack_end (GTK_BOX (priv->vbox_view), priv->find_bar, FALSE, FALSE, 0);
     g_object_set (priv->find_bar, "height-request", -1, NULL);
     g_object_set (priv->find_entry, "height-request", -1, NULL);
     gtk_widget_show_all (priv->find_bar);
@@ -1414,6 +1723,8 @@ view_loaded (YelpView   *view,
     g_free (icon);
     g_free (title);
 
+    app_read_later_changed (priv->application, doc_uri, window);
+
     completion = (GtkTreeModel *) g_hash_table_lookup (completions, doc_uri);
     if (completion == NULL) {
         GtkListStore *base = gtk_list_store_new (5,
@@ -1674,3 +1985,204 @@ hidden_key_press (GtkWidget    *widget,
     }
     return FALSE;
 }
+
+static gboolean
+bookmarks_closed (GtkWindow   *bookmarks,
+                  GdkEvent    *event,
+                  YelpWindow  *window)
+{
+    YelpWindowPrivate *priv = GET_PRIV (window);
+
+    gtk_widget_destroy (GTK_WIDGET (bookmarks));
+    priv->bookmarks_editor = NULL;
+
+    return TRUE;
+}
+
+static void
+bookmarks_set_bookmarks (YelpWindow *window)
+{
+    GVariant *value;
+    GVariantIter *viter;
+    YelpUri *uri;
+    gchar *doc_uri;
+    GSList *entries = NULL;
+    gchar *page_id, *icon, *title; /* do not free */
+    YelpWindowPrivate *priv = GET_PRIV (window);
+
+    gtk_list_store_clear (priv->bookmarks_store);
+
+    g_object_get (priv->view, "yelp-uri", &uri, NULL);
+    doc_uri = yelp_uri_get_document_uri (uri);
+    value = yelp_application_get_bookmarks (priv->application, doc_uri);
+    g_free (doc_uri);
+    g_object_unref (uri);
+
+    g_variant_get (value, "a(sss)", &viter);
+    while (g_variant_iter_loop (viter, "(&s&s&s)", &page_id, &icon, &title)) {
+        YelpMenuEntry *entry = g_new0 (YelpMenuEntry, 1);
+        entry->page_id = page_id;
+        entry->icon = icon;
+        entry->title = title;
+        entries = g_slist_insert_sorted (entries, entry, (GCompareFunc) entry_compare);
+    }
+    for ( ; entries != NULL; entries = g_slist_delete_link (entries, entries)) {
+        GtkTreeIter iter;
+        YelpMenuEntry *entry = (YelpMenuEntry *) entries->data;
+        gtk_list_store_append (priv->bookmarks_store, &iter);
+        gtk_list_store_set (priv->bookmarks_store, &iter,
+                            0, entry->page_id,
+                            1, entry->icon,
+                            2, entry->title,
+                            -1);
+        g_free (entry);
+    }
+    g_variant_iter_free (viter);
+    g_variant_unref (value);
+}
+
+static gboolean
+bookmark_button_press (GtkTreeView    *view,
+                       GdkEventButton *event,
+                       YelpWindow     *window)
+{
+    if (event->button == 3) {
+        GtkTreePath *path = NULL;
+        GtkWidget *menu, *item;
+        GtkTreeSelection *sel = gtk_tree_view_get_selection (view);
+
+        gtk_tree_view_get_path_at_pos (view,
+                                       event->x, event->y,
+                                       &path, NULL,
+                                       NULL, NULL);
+        gtk_tree_selection_select_path (sel, path);
+        gtk_tree_path_free (path);
+
+        menu = gtk_menu_new ();
+        g_object_ref_sink (menu);
+
+        item = gtk_menu_item_new_with_mnemonic (_("_Open Bookmark"));
+        g_object_set_data ((GObject *) item, "new-window", (gpointer) FALSE);
+        g_signal_connect (item, "activate",
+                          G_CALLBACK (bookmark_opened), window);
+        gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
+
+        item = gtk_menu_item_new_with_mnemonic (_("Open Bookmark in New _Window"));
+        g_object_set_data ((GObject *) item, "new-window", (gpointer) TRUE);
+        g_signal_connect (item, "activate",
+                          G_CALLBACK (bookmark_opened), window);
+        gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
+
+        item = gtk_separator_menu_item_new ();
+        gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
+
+        item = gtk_menu_item_new_with_mnemonic (_("_Remove Bookmark"));
+        g_signal_connect_swapped (item, "activate",
+                                  G_CALLBACK (bookmark_remove), window);
+        gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
+
+        gtk_widget_show_all (menu);
+        gtk_menu_popup (GTK_MENU (menu),
+                        NULL, NULL, NULL, NULL,
+                        event->button,
+                        event->time);
+
+        g_object_unref (menu);
+        return TRUE;
+    }
+    return FALSE;
+}
+
+static void
+bookmark_opened (GtkMenuItem *item,
+                 YelpWindow  *window)
+{
+    YelpWindowPrivate *priv = GET_PRIV (window);
+    GtkTreeIter iter;
+    GtkTreeSelection *sel = gtk_tree_view_get_selection (GTK_TREE_VIEW (priv->bookmarks_list));
+
+    if (gtk_tree_selection_get_selected (sel, NULL, &iter)) {
+        YelpUri *base, *uri;
+        gchar *page_id, *xref;
+        gtk_tree_model_get (GTK_TREE_MODEL (priv->bookmarks_store), &iter,
+                            0, &page_id,
+                            -1);
+        xref = g_strconcat ("xref:", page_id, NULL);
+        g_object_get (priv->view, "yelp-uri", &base, NULL);
+        uri = yelp_uri_new_relative (base, xref);
+
+        if (g_object_get_data ((GObject *) item, "new-window"))
+            yelp_application_new_window_uri (priv->application, uri);
+        else
+            yelp_view_load_uri (priv->view, uri);
+
+        g_object_unref (base);
+        g_object_unref (uri);
+        g_free (page_id);
+        g_free (xref);
+    }
+}
+
+static void
+bookmark_activated (GtkTreeView        *view,
+                    GtkTreePath        *path,
+                    GtkTreeViewColumn  *column,
+                    YelpWindow         *window)
+{
+    GtkTreeIter iter;
+    YelpWindowPrivate *priv = GET_PRIV (window);
+
+    if (gtk_tree_model_get_iter (GTK_TREE_MODEL (priv->bookmarks_store),
+                                 &iter, path)) {
+        YelpUri *base, *uri;
+        gchar *page_id, *xref;
+        gtk_tree_model_get (GTK_TREE_MODEL (priv->bookmarks_store), &iter,
+                            0, &page_id,
+                            -1);
+        xref = g_strconcat ("xref:", page_id, NULL);
+        g_object_get (priv->view, "yelp-uri", &base, NULL);
+        uri = yelp_uri_new_relative (base, xref);
+
+        yelp_view_load_uri (priv->view, uri);
+
+        g_object_unref (base);
+        g_object_unref (uri);
+        g_free (xref);
+        g_free (page_id);
+    }
+}
+
+static gboolean
+bookmark_key_release (GtkTreeView *view,
+                      GdkEventKey *event,
+                      YelpWindow  *window)
+{
+    if (event->keyval == GDK_Delete) {
+        bookmark_remove (window);
+        return TRUE;
+    }
+
+    return FALSE;
+}
+
+static void
+bookmark_remove (YelpWindow  *window)
+{
+    YelpWindowPrivate *priv = GET_PRIV (window);
+    GtkTreeIter iter;
+    GtkTreeSelection *sel = gtk_tree_view_get_selection (GTK_TREE_VIEW (priv->bookmarks_list));
+
+    if (gtk_tree_selection_get_selected (sel, NULL, &iter)) {
+        YelpUri *uri;
+        gchar *doc_uri, *page_id;
+        gtk_tree_model_get (GTK_TREE_MODEL (priv->bookmarks_store), &iter,
+                            0, &page_id,
+                            -1);
+        g_object_get (priv->view, "yelp-uri", &uri, NULL);
+        doc_uri = yelp_uri_get_document_uri (uri);
+        yelp_application_remove_bookmark (priv->application, doc_uri, page_id);
+        g_object_unref (uri);
+        g_free (doc_uri);
+        g_free (page_id);
+    }
+}
